/*==========================================================
 * Program : pdu.c                         Project : smslink
 * Authors : Philippe Andersson.
 *           Franois Baligant <francois@euronet.be>.
 * Date    : 08/05/00
 * Version : 0.04a
 * Notice  : (c) Les Ateliers du Heron, 1998 for Scitex Europe, S.A.
 * Comment : Library of PDU-related functions for the smslink servers.
 *
 * Modification History :
 * - 0.01a (07/04/00) : Initial release.
 * - 0.02a (13/04/00) : Added the ascii7topacked8() function.
 * - 0.03a (26/04/00) : Fixed a misunderstanding in encode_pdu().
 *   TP_UDL should be measured in chars, not in bytes. Thanks
 *   to <eric.guidi@wavecom.fr> for pointing that out.
 * - 0.04a (08/05/00) : Fixed a second misunderstanding (pad
 *   the TP_DA before byteswap, not after. Thanks again
 *   to <eric.guidi@wavecom.fr> for pointing that out.
 *========================================================*/

#include <unistd.h>
#include <stdio.h>                         /* for fprintf */
#include <stdlib.h>                  /* for errno & stuff */
#include <errno.h>
#include <string.h>
#include <syslog.h>
#include <math.h>                            /* for pow() */

#include "sms_serv.h"
#include "pdu.h"

/*========================================================*/
/* For debugging purposes only - comment out for normal compile */
/* #define INCL_DEBUG_CODE */

/*========================================================*/
inline int dapad (char *string)
{
  int len;
  
  len = strlen (string);
  if ((strlen (string) % 2) == 1) {
    /* str_insert (string, (len - 1), 'F'); */
    strcat (string, "F");
  }
}                                             /* dapad () */
/*========================================================*/
inline int byteswap (char *string)
{
  int len, i;
  char c;
  
  len = strlen (string);
  for (i = 1; i < len; i += 2) {
    c = string[i];
    string[i] = string[i - 1];
    string[i - 1] = c;
  }
}                                          /* byteswap () */
/*========================================================*/
int ascii7topacked8 (char *in, char *out)
{
  int i, inbitlen, outbitlen, outfree, chunklen;
  unsigned char cur_c, out_c, mask, trans;
  char scratch[3];

  /*--------------------------------------Initializations */
  i = 0;
  outbitlen = 0;
  out_c = '\0';

  /*---------------------------------Main processing loop */
  while (in[i]) {
    cur_c = in[i];
    inbitlen = 7;

    while (inbitlen) {
      outfree = (8 - outbitlen);
      chunklen = min (outfree, inbitlen);
      if (!chunklen) {
        /* out_c is full - save it */
        sprintf (scratch, "%0.2X", out_c);
        strcat (out, scratch);
        /* re-initialize out_c */
        out_c = '\0';
        outbitlen = 0;
        chunklen = inbitlen;
      }                                 /* if (!chunklen) */
      /* save chunklen bits to trans */
      mask = (pow (2, chunklen) - 1);
      trans = cur_c & mask;
      /* offset what's left of cur_c for further use */
      cur_c = cur_c >> chunklen;
      /* offset trans for merge with out_c */
      trans = trans << outbitlen;
      /* merge trans in out_c */
      out_c |= trans;
      /* update length counters */
      inbitlen -= chunklen;
      outbitlen += chunklen;
    }                                 /* while (inbitlen) */

    i++;
  }                                      /* while (in[i]) */
  
  /*------------------------------------------Conclusions */
  /* save the remaining last bits */
  sprintf (scratch, "%0.2X", out_c);
  strcat (out, scratch);

#ifdef INCL_DEBUG_CODE
  fprintf (stderr, "Message in packed8 format : [%s]\n", out);
#endif
  return (strlen (out) / 2);
}                                   /* ascii7topacked8 () */
/*========================================================*/

/*========================================================*/
/*#################### main PDU functions ################*/
/*========================================================*/
char *encode_pdu (struct symbols *symbols)
{
  /* This function is largely based on proof-of-concept code *
   * provided by Franois Baligant <francois@euronet.be>.    */
   
  char *pdu;
  char scratch[MAXPHNUMLEN + 1];
  unsigned int tpdu;
  int skip;
  char numplan[3];
  char dalen[3];
  char *buff;
  int tp_udl;
  
  /*---------------------------------------Initialization */
  pdu = (char *) malloc ((BUFFSIZE + 1) * sizeof (char));
  if (!pdu) {
    syslog ((FACILITY | LOG_ERR), "can't malloc() for PDU string.");
    return (NULL);
  }
  pdu[0] = '\0';
  
  /*--------------------------------------Field #1 : TPDU */
  tpdu = (TP_MTI_SMS_SUBMIT | TP_MMS_NO_MORE_MSG | TP_VPF_RELATIVE | TP_SRI_SR_NOT_REQD);
  /* tpdu = (TP_MTI_SMS_SUBMIT | TP_MMS_NO_MORE_MSG); */
  sprintf (scratch, "%0.2X", tpdu);
  strcat (pdu, scratch);
  
  /*----------------------------------------Field #2 : MR */
  /* Message Reference left NULL - should be set by the phone */
  strcat (pdu, "00");

  /*----------------------------------------Field #3 : DA */
  /* Destination Address : Address Length, Numbering Plan, Number */
  if (symbols->destgsm[0] == '+') {
    /* international number */
    skip = 1;
    strcpy (numplan, "91");
  }
  else if ((symbols->destgsm[0] == '0') && (symbols->destgsm[1] == '0')) {
    /* international number */
    skip = 2;
    strcpy (numplan, "91");
  }
  else {
    /* local number */
    skip = 0;
    strcpy (numplan, "81");
  }
  
  /* remove the international call prefix - if any */
  strcpy (scratch, symbols->destgsm);
  if (skip > 0) {
    shiftleft (scratch, skip);
  }
  
  /* Now store address length as hex value */
  sprintf (dalen, "%0.2X", strlen (scratch));
  strcat (pdu, dalen);
  
  /* Add the numbering plan */
  strcat (pdu, numplan);
  
  /* Finally append the padded, byte-swapped, destination */
  dapad (scratch);
  byteswap (scratch);
  strcat (pdu, scratch);

  /*---------------------------------------Field #4 : PID */
  sprintf (scratch, "%0.2X", PID_SMS);
  strcat (pdu, scratch);
  
  /*---------------------------------------Field #5 : DCS */
  sprintf (scratch, "%0.2X", DCS_PCCP437);
  strcat (pdu, scratch);
  
  /*----------------------------------------Field #6 : VP */
  if (tpdu & TP_VPF_RELATIVE) {
    strcat (pdu, "AA");                         /* 4 days */
  }
  
  /*----------------------------Field #7 : TP_UDL + TP_UD */
  /* TP_UDL is the length of TP_UD measured in chars (NOT in  */
  /* bytes). Thanks to  <eric.guidi@wavecom.fr>).             */
  buff = (char *) malloc (((2 * MAXMSGLEN) + 1) * sizeof (char));
  buff[0] = '\0';
  
  tp_udl = strlen (symbols->message);
  ascii7topacked8 (symbols->message, buff);

  sprintf (scratch, "%0.2X", tp_udl);
  strcat (pdu, scratch);
  strcat (pdu, buff);
  /*-------------------------------------------Conclusion */
#ifdef INCL_DEBUG_CODE
  fprintf (stderr, "PDU-encoded string : [%s]\n", pdu);
#endif
  free (buff);
  return (pdu);
}                                        /* encode_pdu () */
/*==========================================================
 * EOF : pdu.c
 *===================*/
